%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:           PXC:Mips-spec.sl
% Title:          Some special things for MIPS PSL compiler
% Author:         Winfried Neun , ZIB Berlin
% Created:        September 15, 1989
% Modified:
% Status:         Experimental
% Mode:           Lisp
%
% Copyright (c) 1989, Konrad-Zuse-Zentrum Berin, All Rights Reserved
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% a special Linke to save the NOP instruction after final Jump

(de *LinkE (DeAllocCount FunctionName FunctionType NumberOfArguments)
  (cond ((FlagP FunctionName 'ForeignFunction)
              `((*ForeignLink ,FunctionName ,FunctionType ,NumberOfArguments) 
                (*Exit 0)))
        ((get functionname 'exitopencode)
              `((*DeAlloc ,DeAllocCount)
                (*JCall ,FunctionName)))
        ((get functionname 'opencode)
              `((*Call ,FunctionName)
                (*exit ,DeAllocCount)))
        ((InternallyCallableP functionname)
              `((*dealloc ,Dealloccount)
		(J (InternalEntry ,functionname))
                (nop)))
        (t    `((*dealloc ,Dealloccount)
                (*JCall ,functionname)))))

% change branch to exitlabel into *exit when framesize ne 0 too

(fluid '(exitt&))

(de &jumpopt (u)
    (when (equal (cadar u) exitt&) (rplaca u '(*exit (*framesize)))))

% turn off lapopt in faslend

(fluid '(*lapopt))

(setq *lapopt t)

(when (funboundp 'old-faslend) (copyd 'old-faslend 'faslend))

(de faslend () 
    (prog (*lapopt)
       (return (old-faslend))))

% open codes for cons ncons xcons

(put 'cons 'opencode
      '((lui (reg 5) 16#4000)             % pair-tag
        (sll (reg t5) (reg r16) 2) 
        (sw (reg 1) (indexed (reg t5) 0)) % Always sufficient memory for
        (sw (reg 2) (indexed (reg t5) 4))
        (or (reg 1) (reg 5) (reg r16)) % the result
        (*wplus2 (reg r16) 2)
        (*jumpwlessp (labelgen templabel) (reg r16) (reg r17)) 
        (*savefastframe)
        (*link !%reclaim expr 0)
        (*restorefastframe)  
        (labelref templabel))
)
(put 'xcons 'opencode
      '((lui (reg 5) 16#4000)             % pair-tag
        (sll (reg t5) (reg r16) 2) 
        (sw (reg 2) (indexed (reg t5) 0)) % Always sufficient memory for
        (sw (reg 1) (indexed (reg t5) 4))
        (or (reg 1) (reg 5) (reg r16)) % the result
        (*wplus2 (reg r16) 2)
        (*jumpwlessp (labelgen templabel) (reg r16) (reg r17)) 
        (*savefastframe) 
        (*link !%reclaim expr 0)
        (*restorefastframe)  
        (labelref templabel))
)
(put 'ncons 'opencode
      '((lui (reg 5) 16#4000)             % pair-tag
        (sll (reg t5) (reg r16) 2) 
        (sw (reg 1) (indexed (reg t5) 0)) % Always sufficient memory for
        (sw (reg nil) (indexed (reg t5) 4))
        (or (reg 1) (reg 5) (reg r16)) % the result
        (*wplus2 (reg r16) 2)
        (*jumpwlessp (labelgen templabel) (reg r16) (reg r17)) 
        (*savefastframe) 
        (*link !%reclaim expr 0)
        (*restorefastframe)       
        (labelref templabel))
)

(defcmacro *savefastframe)

(de  *savefastframe ()
   (cond ((null *usefastframe) '((*xalloc 1) (*move (reg 1) (xframe 1)) ))
         (T (append `((*xalloc ,(wplus2 nalloc* 1)) (*move (reg 1) (xframe 1)))
                     (pnth '((*move (reg ff4) (xframe 5))
                             (*move (reg ff3) (xframe 4))
                             (*move (reg ff2) (xframe 3)) 
                             (*move (reg ff1) (xframe 2))
                             (*move (reg nil) (reg nil))) % dummy for reduce
                         (wdifference 5 nalloc*))))))

(defcmacro *restorefastframe)

(de  *restorefastframe () 
   (cond ((null *usefastframe) '((*move (xframe 1) (reg 1))  (*xdealloc 1)))
        (T (append   (pnth '((*move (xframe 5) (reg ff4)) 
                             (*move (xframe 4) (reg ff3)) 
                             (*move (xframe 3) (reg ff2))  
                             (*move (xframe 2) (reg ff1))
                             (*move (reg nil) (reg nil))) % dummy for reduce 
                         (wdifference 5 nalloc*))
           `((*move (xframe 1) (reg 1)) (*xdealloc ,(wplus2 nalloc* 1))))) ))
 
(commentoutcode
(put 'bittable 'opencode
      '((srl (reg 3) (reg 2) 2)         % offset
        (sll (reg 4) (reg 1) 2)
        (add (reg 3) (reg 3) (reg 4))
        (lbu (reg 5) (displacement (reg 3) 0))
        (andi (reg 4) (reg 2) 3)
        (sll (reg 4) (reg 4) 1)
        (*move 6 (reg 2))
        (sub (reg 4) (reg 4) (reg 2))   %   fraction
        (srlv (reg 1) (reg 5) (reg 4))
        (andi (reg 1) (reg 1) 3)))
 
(put 'putbittable 'opencode
      '((srl (reg t3) (reg 2) 2) % offset
        (sll (reg 4) (reg 1) 2)
        (Add (reg t3) (reg 4) (reg t3))
        (lbu (reg 5) (displacement (reg t3) 0) (reg 5))
        (andi (reg 4) (reg 2) 3)   % %4
        (*move 3 (reg T2))
        (sub (reg 4) (reg t2) (reg 4))  % 3 - (pos%4)
        (sll (reg 4) (reg 4) 1)         % *2 / =fraction
        (sllv (reg T2) (reg T2) (reg 4))
        (nor (reg t2) (reg zero) (reg t2))
        (and (reg 5) (reg T2) (reg 5))
        (andi (reg 3) (reg 3) 3)       % value & 3
        (sllv (reg t1) (reg 3) (reg 4)) % <<fraction
        (or (reg 5) (reg T1) (reg 5))  % or
        (sb (reg 5) (displacement (reg T3) 0))
        (*move (reg 3) (reg 1))))
)

(put 'bittable 'opencode
      '((srl (reg 3) (reg 2) 2) % offset
        (sll (reg 4) (reg 1) 2)
        (add (reg 3) (reg 3) (reg 4))
        (lbu (reg 5) (displacement (reg 3) 0))
        (andi (reg 4) (reg 2) 3)
        (sll (reg 4) (reg 4) 1)
        (*move 6 (reg 2))
        (sub (reg 4) (reg 2) (reg 4))   % - fraction
        %(sub (reg 4) (reg 0) (reg 4))   %   fraction
        (srlv (reg 1) (reg 5) (reg 4))
        (andi (reg 1) (reg 1) 3)))
 
(put 'putbittable 'opencode
      '((srl (reg t3) (reg 2) 2) % offset
        (sll (reg 4) (reg 1) 2)
        (Add (reg t3) (reg 4) (reg t3))
        (lbu (reg 5) (displacement (reg t3) 0))
        (andi (reg 4) (reg 2) 3)   % %4
        (*move 3 (reg T2))
        (sub (reg 4) (reg t2) (reg 4))  % 3 - (pos%4)
        (sll (reg 4) (reg 4) 1)         % *2 / =fraction
        (sllv (reg T2) (reg T2) (reg 4))
        (nor (reg t2) (reg null) (reg t2))
        (and (reg 5) (reg T2) (reg 5))
        (andi (reg 3) (reg 3) 3)       % value & 3
        (sllv (reg t1) (reg 3) (reg 4)) % <<fraction
        (or (reg 5) (reg T1) (reg 5))  % or
        (sb (reg 5) (displacement (reg T3) 0))
        (*move (reg 3) (reg 1))))
 

% This code is intended for filling the 'holes' in the frame.
% e.g. in the following function
%
%  (de hugo (x) (cond ((null x) (call-another))
%                     (t (prog (a b c) ....))
%
% a frame of length 4 will be allocated, but it will not be preset (by nil)
% completely at entering the functiont but the frmae elements for a b and c
% will be set (to nil) if the prog is entered.
% So if the (null x) case is taken we have 3 unset frame elements on the stack
% which contain random information and this is dangerous for gc.
% to resolve that problem, the following function will work on the input
% to pass-1-lap and insert nil settings where needed.

(fluid '(&fillframeholes &frame-numbers))

(compiletime (flag '(findcmacro lookatinstruction) 'internalfunction)) 
 
(de &fillframeholes (code)
   (prog (allo cmactoextern externfound ll lll nn linearcode restcode)
      (when (not &fillframeholes) (return code))
      (setq cmactoextern
                  '(*link *linke *call *jcall *cons *xcons *fastmkfltn))
             % trivial tests whether we have to work at all
      (unless (setq allo (findcmacro code '*alloc)) (return code))
      (setq ll cmactoextern)
      (while (and (not externfound) ll)
            (setq externfound (findcmacro code (car ll))) 
            (setq ll (cdr ll)))  
      (when (not externfound) (return code))
 
            % so we have to work, starting at the *alloc 
 
      (setq nn (cadar allo))  %the frmae size              
      (setq linearcode t)
      (setq restcode (cdr allo))       
 
      (setq &frame-numbers nil) 
 
      (while (and linearcode restcode)
         (setq linearcode (lookatinstruction (car restcode)))
         (setq restcode (cdr restcode))) 
 
      (when linearcode (return code))
 
      (setq lll (setq ll (cons nil nil))) 
              
             % search frame that have not been set, prepare code,
             % put into place after *alloc
 
      (for (from i 1 nn 1) (do
            (unless (memq i &frame-numbers)
                  (rplaca ll `(*move (reg nil) (frame ,i)))
                  (when (eq &fillframeholes 'verbose) 
                    (prin2 "++ in ") (prin2 name&)
                    (prin2 " presetting frame ")
                    (prin2t i))
                  (push 'nil ll)))) 
      
      (when (eq lll ll) (return code))
      (rplacd lll (cdr allo))
 
      (rplacd allo (cdr ll))
      (return code))) 
         
(de lookatinstruction (instr)
      (cond ((atom instr) t)
            ((and (eq (car instr) '*move) 
                  (eqcar (caddr instr) 'frame)) 
                       (push (cadr (caddr instr)) &frame-numbers) 17)
            ((memq (car instr) '(*link *call *linke *jcall *jumpintype
                 *jumpeq *jump  *JumpType *JumpNotType *JumpNotInType
                 *jumpnoteq *jumpwlessp *jumpwgreaterp *jumpwleq *jumpwgeq
                 *jumpon)) nil)
            (t t)))
 
(de findcmacro (list pat) 
     (cond ((atom list ) nil) 
           ((eqcar (car list) pat) list) 
           (t        (findcmacro (cdr list) pat )))) 

(setq &fillframeholes t)
