%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% WINHELP1.red    shell for REDUCE help under windows  
%                 maps help I/O to help server in separate task
%
%  Herbert Melenk, Jan 1992
%
%  RLISP
%


fluid '(!*help!-handle!*);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  1. low level interface: connecting to help server
%

symbolic procedure win!-open!-help();
  % Try to open or reopen connection resp. start help server.
   (!*help!-handle!*
          and
    fetch!-server(!*help!-handle!*,"HANDSHAKE")
          and
    !*help!-handle!*)
 or
    win!-open!-help1()
 or
    (system "$reduce\exe\helpserv.exe"
          and
     win!-open!-help1())
  ;
        
procedure win!-open!-help1();
  !*help!-handle!* := 
       open!-server("ReduceHelp","Help",'win!-help!-shell);
 
procedure win!-send!-help(l);
  begin scalar command;  
    command := car l; l := cdr l;
    command := if idp command then id2string command else
                        bldmsg("%w", command);
    send!-server (!*help!-handle!*, command, bldmsg( "%l", l));  
  end;

procedure win!-fetch!-help(command);
  fetch!-server(!*help!-handle!*, bldmsg("%l", command));

procedure win!-close!-help();
  <<close!-server  !*help!-handle!*;
    !*help!-handle!* := nil>>;


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% medium layer: encapsulate help output by a memory file
%

fluid '(!*help!-initialized);

symbolic procedure win!-help();
  % this function is called internally when the HELP menu in
  % the main REDUCE window is activated.
  % Initializes local HELP system and HELP server.
 begin 
   if not !*help!-initialized then helpinit("help.hlp","help.idx");
   if null win!-open!-help() then
     rederr "WINDOWS help server not available";
   if lastpname!* then win!-help!-apply('h!/error,nil);
 end;
 
symbolic procedure win!-help!-shell(mode,text);
  % asynchronous entry for the help system.
  %   mode   Id    : BUTTON
  %                  TOPIC
  %                   ...
  %   text    string
  %       
    if mode='BUTTON then 
       win!-help!-apply(intern text,nil) 
      else
       win!-help!-error bldmsg("undefined HELP mode >%w<",mode);

symbolic procedure win!-help!-apply(fcn,args);
     % apply a help function, but collecting
     % the output in a list of strings  and send
     % them to the help server.
  begin scalar f,of,r,l;
    if null getd fcn then return 
      win!-help!-error bldmsg("undefined HELP function >%w<",fcn);
    f := mf!-open("hugo",'OUTPUT);
    of := wrs f;   
    r := errorset(fcn. for each a in args collect mkquote a,nil,nil);
    if atom r then prin2t "error during help apply";
    l := mf2list(f);
    wrs of;
    close f;
    win!-send!-help{'begin_transaction};
    win!-send!-help{'clear,'TEXT,'TEXT};
    for i:=1:length l do
      win!-send!-help{'putline,'TEXT,i,strip!-txtline nth(l,i)};
    win!-send!-help{'end_transaction};
  end;
 
fluid '(!*htable);

symbolic procedure help0();
   begin scalar s,tp;
     % fetch the argument for help;
      s:= win!-fetch!-help '(topic);
      if s then
        h!/topics(htable!*,list intern s)
          else h!/error();
   end;

symbolic procedure win!-help!-error(u);
   win!-help!-apply('prin2t,{u});

                                           
symbolic procedure strip!-txtline txt;
  % this is a destructive routine and very PSL specific.
  begin integer l,c;
    l := strlen strinf txt;
    for i:=l step -1 until 0 do
      if (c:=strbyt(strinf txt,i)) = 10 or c=13 then l:=i;
    setf(strbyt(strinf txt,l),0); 
    return copy!-message txt;  % make a PSL string of correct length
  end;
end;        
