%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PNK:ALLOCATORS.SL
% Title:        Low level storage management
% Author:       Eric Benson
% Created:      27 August 1981
% Modified:     03-Jul-84 10:14:03
% Status:       Experimental
% Mode:         Lisp
% Package:      Kernel
% Compiletime:
% Runtime:
%
% (c) Copyright 1983, 1984, Hewlett-Packard Company, all rights reserved.
% (c) Copyright 1981, University of Utah
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 1-Jun-88 (Neun ZIB) fixed get-heap-trap ti honour heaptrapbound
% 18-Apr-86 (Melenk ZIB) moved DelBps to Kernel in favour of faslin
% 21-Jan-86 (Melenk ZIB) variable str renamed to st (CRAY)
% 03-Jul-84 10:09:24 (Jim Ambras/CRC)
%  Modified get-heap-try-reclaim to use stderror instead of kernel-fatal-error
%   so the kernel won't terminate when heap space is exhausted.
% 23-Feb-84 15:40:14 (Brian Beach)
%  Removed everything that is in the kernel.  (This file will now be FASLed in)
% 21-Feb-84 09:52:00 (Steve Lowder)
%  Removed all useage of wvar, wconst, warray, wstring.
% 03-Jan-84 18:25:44 (Cris Perdue)
%  Fixed some comments, made gtbps grab no space if the block
%  requested is too big to give.
% 02-Dec-83 09:32:39 (Brian Beach)
%   Translated from Rlisp to Lisp.
%
% 5 Nov 83, Swanson and Perdue
%  Changed mention of PDP10 to mention the new system attribute
%  OneSpaceGC.  Added special code for Extended 20.
%  <PSL.KERNEL>ALLOCATORS.RED.7, 23-Mar-83 11:35:37, Edit by KESSLER
%  Added OldHeapTrapBound to exported WVars, so we can update the heap trap
%  bound upon switch.
% Edit by Cris Perdue, 16 Feb 1983 1834-PST
% Pre-GC trap, known-free-space fns
%  <PSL.KERNEL>ALLOCATORS.RED.4, 10-Jan-83 15:50:50, Edit by PERDUE
%  Added GtEVect
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(fluid '(heapupperbound
       heaplowerbound
       heaplast
       heaptrapbound
         heaplast
       heaptrapped
       nextbps
       lastbps))
 
(setf heaptrapped nil)
 
(on fast-integers)
 
(de known-free-space ()
  (mkint (wquotient (wdifference heapupperbound heaplast)
                    addressingunitsperitem)))
 
(de free-bps ()
  (wquotient (wdifference lastbps nextbps)
           addressingunitsperitem)
  )
%
% The function GTHEAP is defined in the kernel.  It calls GET-HEAP-TRAP when
% HEAPLAST (the pointer to the next available item) passes HEAPTRAPBOUND.
%
 
(de get-heap-trap (pointer number-of-items)
 
  % An attempt has been made to allocate NUMBER-OF-ITEMS from the heap, and
  % the HEAPTRAPBOUND was exceeded.  POINTER is the value that would have been
  % returned if the trap bound had not been passed.  The HEAPTRAPPED flag is used
  % to indicate when GC-TRAP has already been called.  If it has not, do so.
 
  (if (null heaptrapped)
    (progn
     (setf heaplast (wdifference heaplast (wtimes2 number-of-items
                                       addressingunitsperitem)))
     (setf heaptrapped t)
     (gc-trap)
     (gtheap number-of-items)
     )
 %% (if (wlessp heaplast heapupperbound)
    (if (wlessp heaplast heaptrapbound)
      pointer
      (get-heap-try-reclaim number-of-items)
      )
    ))
 
(de gc-trap ()
  )
 
(de get-heap-try-reclaim (number-of-items)
  % Reset HEAPLAST to the way it was before the allocation attempt, do
  % a garbage collection, and then try the allocation again.
 
  (setf heaplast (wdifference heaplast (wtimes2 number-of-items
                                    addressingunitsperitem)))
  (reclaim)
  (let ((result heaplast))
    (setf heaplast (wplus2 heaplast (wtimes2 number-of-items
                                   addressingunitsperitem)))
    (if (wlessp heaplast heaptrapbound)
      result
      (progn
       (setq heaplast result)              % Reset heaplast
       (if (add-heap-space number-of-items) (gtheap number-of-items)
                  (stderror "Ran out of heap space.")))
      )))
 
(de gc-trap-level ()
  (wquotient (wdifference heapupperbound heaptrapbound)
           addressingunitsperitem))
 
(de set-gc-trap-level (n)
  (unless (intp n)
    (nonintegererror n 'set-gc-trap-level))
  (setq n (intinf n))
  (setq heaptrapbound
        (wdifference heapupperbound
                     (wtimes2 n addressingunitsperitem)))
  t)
 
(de delheap (lowpointer highpointer)
  (when (weq highpointer heaplast)
    (setq heaplast lowpointer)))
 
%(de gtstr (upper-bound)        Defined in kernel
  % Allocate a string of UPPER-BOUND+1 characters.
%  (let* ((n-words  (strpack upper-bound))
%       (st       (gtheap (wplus2 1 n-words))))
%    (setf (getmem st ) (mkitem hbytes-tag upper-bound))
%   (setf (wgetv st  n-words) 0)  % clear last word, including last byte
%   st
%   ))
 
% GTCONSTSTR is defined in the kernel
 
(de gthalfwords (upper-bound)
  % Allocate space for a halfwords vector of UPPER-BOUND+1 elements.
  (let* ((n-words  (halfwordpack upper-bound))
       (ptr      (gtheap (wplus2 n-words 1))))
    (setf (getmem ptr) (mkitem hhalfwords-tag upper-bound))
    ptr
    ))
 
(de gtvect (upper-bound)
  % Allocate space for a vector of UPPER-BOUND+1 elements.
  (let ((ptr  (gtheap (wplus2 (vectpack upper-bound) 1))))
    (setf (getmem ptr) (mkitem hvect-tag upper-bound))
    ptr
    ))
 
(de gtevect (upper-bound)
  % Allocate space for an evector of UPPER-BOUND+1 elements.
  (let ((ptr  (gtheap (wplus2 (evectpack upper-bound) 1))))
    (setf (getmem ptr) (mkitem hvect-tag upper-bound))
    ptr
    ))
 
% GTWRDS is defined in the kernel
 
(de gtfixn ()
  % Allocate space for a fixnum.
  (let ((ptr  (gtheap (wplus2 (fixpack) 1))))
    (setf (getmem ptr) (mkitem hwords-tag (wplus2 (fixpack) -1)))
    ptr
    ))
 
(de gtfltn ()
  % Allocate space for a floating point number.
  (let ((ptr  (gtheap (wplus2 (floatpack) 1))))
    (setf (getmem ptr) (mkitem hwords-tag (wplus2 (floatpack) -1)))
    ptr
    ))
 
% GTID is defined in the kernel.
 
% GTBPS and DelBps are defined in the kernel.
 
(de gtwarray (n)
  %. Allocate N words for WVar/WArray/WString
  (if (null n)
    (wquotient (wdifference lastbps nextbps)
             addressingunitsperitem)
    (let ((b (wdifference lastbps (wtimes2 n addressingunitsperitem))))
         (if (and (wgreaterp nextbps b)
                  (not (try-other-bps-spaces n)))
             (stderror '"Ran out of WArray space")
             (mkwrds (setq lastbps
                             (wdifference lastbps n))))
      )))
 
(de delwarray (bottom top)
  %. Return space for WArray
  (when (weq lastbps (inf bottom))
        (setq lastbps (inf top))))
 
 
(off fast-integers)
 
