%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PNK:BACKTRACE.SL
% Title:        Functions for displaying the return stack.
% Author:       ???
% Created:      ???
% Modified:     22-Mar-84 09:23:19 (Brian Beach)
% Status:       Experimental
% Mode:         Lisp
% Package:      Kernel
% Compiletime:
% Runtime:
%
% (c) Copyright 1983, Hewlett-Packard Company, all rights reserved.
%
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 22-Feb-86 added Cray-version of returnaddressp (ZIB Berlin)
%           Fixed calculation of top-of-stack for Cray
%           commentedout verbosebacktrace
% 22-Mar-84 09:13:24 (Brian Beach)
%  Fixed constants.  Removed WFORS.
% 01-Dec-83 14:12:23 (Brian Beach)
%   Translated from Rlisp to Lisp.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(on fast-integers)
 
(compiletime
  (flag '(backtrace1 backtracerange returnaddressp) 'internalfunction))
 
(fluid '(ignoredinbacktrace* options* interpreterfunctions*
         *system-is-fast-linked*))
 
(setq ignoredinbacktrace*
      (list 'eval 'apply 'fastapply 'codeapply 'codeevalapply 'catch
            'codeevalapplyaux 'pre-main
            'errorset 'evprogn 'toploop 'breakeval 'bindeval 'break 'main
%           'stackcheck))
            'top-loop-eval-and-print))
 
(setq interpreterfunctions* (list 'cond 'prog 'and 'or 'progn 'setq))
 
(fluid '(heapupperbound stacklowerbound))
 
(de interpbacktrace ()
  (prog (here)
        (setq here (set-stack-aux))% set-stack-aux defined in memsizes
        (printf "Backtrace, including interpreter functions, from top of stack:%n")
        (return (backtracerange here stacklowerbound 1))))
 
(de backtrace ()
  (prog (here x)
        (setq here (set-stack-aux))
        (printf "Backtrace from top of stack:%n")
        (return (backtracerange here stacklowerbound 0))))
 
(de backtracerange (starting ending interpflag)
  (prog (x memo)
    (when *system-is-fast-linked*
       (print "*** Backtrace for fast-linked PSL ***"))
    (for (from i starting ending -1)
       (do (setq memo (getmem i))
           (cond ( (and (weq (tag memo) btr-tag)
                        (wgreaterp (inf memo) 0)
                        (wlessp (inf memo) maxsymbols))
                     (backtrace1 (mkid (inf memo)) interpflag))
                 ((setq x (returnaddressp memo))
                     (backtrace1 x interpflag)))))
    (return (terpri))))
 
(commentoutcode
(de verbosebacktrace ()
  (prog (here x)
    (unless (member 'addr2id options*)
      (load addr2id))
    (setq here (loc here))
    (for % start a little before here
     (from i here stacklowerbound
         (wminus
          (wtimes2 addressingunitsperitem
                 stackdirection)))
     (do (cond ((and (codep (getmem i))
                 (wgreaterp (inf (getmem i)) heapupperbound))
            (writechar (char tab))
            (channelwriteunknownitem out* (getmem i))
            (terpri))
             ((weq (tag (getmem i)) btr-tag)
            (printf "      %r%n" (mkid (inf (getmem i)))))
             ((setq x (returnaddressp (getmem i)))
            (printf "%p -> %p:%n"
                  (code-address-to-symbol (inf (getmem i))) x))
             (t (printf "      %p%n" (getmem i))))))
    (return (terpri))))
) %commentoutcode
 
(de backtrace1 (item cde)
  %
  % Code is 1 if Interpreter functions should be printed, 0 if not.
  %
  (unless (memq item ignoredinbacktrace*)
    (unless (and (equal cde 0) (memq item interpreterfunctions*))
      (prin1 item)
      (writechar (char blank)))))
 
(off fast-integers)
 
(fluid '(symfnc nextbps))
 
(de returnaddressp (addr)
    (prog (waddr parc1 parc2)
       (setq waddr (wshift addr -2))
       (cond ((or (wgreaterp waddr nextbps) (wlessp waddr 8#200))
                 (return nil)))
       (setq parc1 (craygetparcel (wplus2 addr -2)))
       (setq parc2 (craygetparcel (wplus2 addr -1)))
       (cond ((wneq (wshift parc1 -9) 7)     % R Instruction?
              (return nil)))
%      (putoct addr) (putoct parc1) (putoct parc2) (putoct symfnc)
%      (console-newline)
       (setq parc1 (wand parc1 8#777))
       (setq parc1 (wor (wshift parc1 16) parc2))
       (setq parc1 (wshift parc1 -2))        %    wordaddress
       (setq parc1 (wdifference parc1 symfnc))
       (cond ((or (wlessp parc1 0) (wgreaterp parc1 maxsymbols))
              (return (if (not *system-is-fast-linked*) nil
                          (map-at waddr) nil))))
       (return (mkid parc1))
)   )
