%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PNK:BINDING.SL
% Title:        Primitives to support Lambda binding
% Author:       Eric Benson
% Created:      18 August 1981
% Modified:     23-May-84 10:30:23 (Brian Beach)
% Status:       Experimental
% Mode:         Lisp
% Package:      Kernel
% Compiletime:
% Runtime:
%
% (c) Copyright 1983, Hewlett-Packard Company, all rights reserved.
% (c) Copyright 1981, University of Utah
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
% 30-May-86 W. Neun   , ZIB Berlin
%         added macros to do fluid-to-reg conversion
% 03-Feb-86 H. Melenk , ZIB Berlin
%         added fast bindings and compiletime-load for inum
% 06-Mar-84 17:11:53 (Jim Ambras/CRC)
%  Uppded binding stack size from 4000 to 8000 for REDUCE.
% 02-Mar-84 09:29:17 (Jim Ambras/CRC)
%  Corrected file header.
% 16-Jan-84 14:11:00 (Tim Tillson)
%  Upped binding stack size from 2000 to 4000 for unknown (and possibly buggy?)
%  reasons to build NMODE successfully.
% 01-Dec-83 14:16:18 (Brian Beach)
%   Translated from Rlisp to Lisp.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(fluid '(bndstk))
(fluid '(bndstkptr))
(fluid '(bndstkupperbound))
(fluid ' bndstklowerbound))
 
(compiletime (load inum ))
 
% Binding stack is initialized in the kernel.
% Only the macros BndStkID, BndStkVal and AdjustBndStkPtr will be used
% to access or modify the binding stack and pointer.
 
(de bstackoverflow ()
  (channelprin2 errout* "***** Binding stack overflow, restarting...")
  (channelwritechar errout* (char eol))
  (reset))
 
(de bstackunderflow ()
  (channelprin2 errout* "***** Binding stack underflow, restarting...")
  (channelwritechar errout* (char eol))
  (reset))
 
(de captureenvironment ()
  bndstkptr
  )
 
%
% following macros generates code to do the id-to-reg conversion
% for the bindings
% NOTABENE:     only those fluids are bound, that are flagged
%                      *user-variable
 
(compiletime
 (progn
 
(fluid '(collectedregs*))
 
(de collectregs-read (x)
     (when (and (get x 'registername) (flagp x '*user-variable))
        (setq collectedregs*
         (cons `((eq idname (quote ,x)) ,x) collectedregs*))))
 
(dm idtoreg-conversion-read ()
       (setq collectedregs* nil)
       (mapobl (function collectregs-read))
       (cons 'cond (nconc collectedregs* `((t (symval idnam)))  )))
 
(de collectregs-write (x)
     (when (and (get x 'registername) (flagp x '*user-variable))
      (setq collectedregs*
       (cons `((eq idname (quote ,x)) (setq ,x valuetobind))
                                                     collectedregs*))))
 
(dm idtoreg-conversion-write ()
       (setq collectedregs* nil)
       (mapobl (function collectregs-write))
       (cons 'cond (nconc collectedregs*
         `((t (setf (symval idnam) valuetobind))) )))
 
(de collectregs-rewrite (x)
     (when (and (get x 'registername) (flagp x '*user-variable))
      (setq collectedregs*
       (cons `((eq idname (quote ,x)) (setq ,x (bndstkval bndstkptr)))
                                                     collectedregs*))))
 
(dm idtoreg-conversion-rewrite()
       (setq collectedregs* nil)
       (mapobl (function collectregs-rewrite))
       (cons 'cond (nconc collectedregs*
   `((t (setf (symval (bndstkid bndstkptr))(bndstkval bndstkptr)))) )))
 
)) %end compiletime
 

(de restoreenvironment (ptr)
  % Restore old bindings
 (prog (idname)
  (if (wlessp ptr bndstklowerbound)
    (bstackunderflow)
    (while (wgreaterp bndstkptr ptr)
      (setq idname (mkid (bndstkid bndstkptr)))
      (idtoreg-conversion-rewrite)
%     (setf (symval (bndstkid bndstkptr)) (bndstkval bndstkptr))
      (setq bndstkptr (adjustbndstkptr bndstkptr -1))))))

 
(de clearbindings ()
  %. Restore bindings to top level
  (progn (restoreenvironment bndstklowerbound)
         (!%clear-catch-stack)))
 
(de unbindn (n)
  %. Support for Lambda and Prog interp
  (restoreenvironment (adjustbndstkptr bndstkptr (wminus (intinf n)))))
 
(de lbind1 (idname valuetobind)
  %. Support for Lambda
  (cond ((not (idp idname)) (noniderror idname "binding"))
        ((or (null idname) (weq idname 't))
         (stderror '"T and NIL cannot be rebound"))
        (t (setq bndstkptr (adjustbndstkptr bndstkptr 1)) (if (wgreaterp
                                                               bndstkptr
                                                               bndstkupperbound)
             (bstackoverflow)
              (prog  (idnam)
                (setq idnam (idinf idname))
                (setf (bndstkid bndstkptr) idnam)
                (setf (bndstkval bndstkptr) (idtoreg-conversion-read))
                (idtoreg-conversion-write))))))
 
(de pbind1 (idname)
  %. Support for PROG
  (lbind1 idname nil))
 
(de Pbindn (idlist)   % pbind a list of ids. Used by "prog and friends"
 (prog (n idname idnam valuetobind)
  (setq n 0)
lupe
  (when (not (pairp idlist))  (return n))
  (setq idname (car idlist))
  (setq n (wplus2 n 1))
  (when (not (idp idname)) (return (noniderror idname "binding")))
  (when (or (null idname) (weq idname 't))
                  (return (stderror '"T and NIL cannot be rebound")))
  (setq bndstkptr (adjustbndstkptr bndstkptr 1))
  (when (wgreaterp bndstkptr bndstkupperbound)
                                        (return (bstackoverflow)))
  (setq idnam (idinf idname))
  (setf (bndstkid bndstkptr) idnam)
  (setf (bndstkval bndstkptr) (idtoreg-conversion-read))
  (idtoreg-conversion-write)
  (setq idlist (cdr idlist))
  (go lupe)
))
 
 %%%%%%%%%%%%%%%%%%%%%%%%%%%% fast bindings
%(fluid '(LambindArgs*))
%(setq LambindArgs* (gtwarray 16))
 
(compiletime (setq syslisp nil))
 (lap '((*entry fill-lambindargs expr 1)
        (*alloc 0)
        (*move (reg 1) ($fluid lambindargs*))
        (*exit 0)
)      ) 
(compiletime (setq syslisp t))
 
(fill-lambindargs LamBindArgs*)
 
(de lambind (v)
   (prog (n)
      (setq v (vecinf v))
      (setq n (veclen v))
      (ifor (from i 0 n 1)
         (do(lbind1 (vecitm v i) (wgetv (vecinf LambindArgs*) i)))
)  )  )
 
 
(de progbind (v)
   (prog (n)
      (setq v (vecinf v))
      (setq n (veclen v))
      (ifor (from i 0 n 1)
         (do(pbind1 (vecitm v i)))
)  )  )
