%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PNK:BACKTRACE.SL 
% Title:        Functions for displaying the return stack. 
% Author:       ??? 
% Created:      ??? 
% Modified:     22-Mar-84 09:23:19 (Brian Beach) 
% Package:      Kernel 
%
% (c) Copyright 1987, University of Utah, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 27-May-87 (Harold Carr & Leigh Stoller)
%  Added fluid declaration of in* and out*.
% 22-Mar-84 09:13:24 (Brian Beach)
%  Fixed constants.  Removed WFORS.
% 01-Dec-83 14:12:23 (Brian Beach)
%   Translated from Rlisp to Lisp.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(on fast-integers)

(compiletime
  (flag '(backtrace1 backtracerange) 'internalfunction))

(fluid '(ignoredinbacktrace* options* interpreterfunctions*))
(fluid '(in* out*))

(setq ignoredinbacktrace*
      (list 'eval 'apply 'fastapply 'codeapply 'codeevalapply 'catch
            'errorset 'evprogn 'toploop 'breakeval 'bindeval 'break 'main
            'stackcheck))

(setq interpreterfunctions* (list 'cond 'prog 'and 'or 'progn 'setq))

(global '(heapupperbound stacklowerbound))

(de interpbacktrace ()
  (prog (here)
        (setq here (loc here))
        (printf "Backtrace, including interpreter functions, from top of stack:%n")
        (return (backtracerange here stacklowerbound 1))))

(de backtrace ()
  (prog (here x)
        (setq here (loc here))
        (printf "Backtrace from top of stack:%n")
        (return (backtracerange here stacklowerbound 0))))

(de backtracerange (starting ending interpflag)
  (prog (x)
    (for (from i starting ending
               (wminus
                (wtimes2 addressingunitsperitem
                         stackdirection)))
	 (do (cond ((weq (tag (*get-stack i)) btr-tag)
		    (backtrace1 (mkid (inf (*get-stack i))) interpflag))
		   ((setq x (returnaddressp (*get-stack i)))
                    (backtrace1 x interpflag)))))
    (return (terpri))))

(de verbosebacktrace ()
  (prog (here x)
    (unless (member 'addr2id options*)
      (load addr2id))
    (setq here (loc here))
    (for % start a little before here
     (from i here stacklowerbound
           (wminus
            (wtimes2 addressingunitsperitem
                     stackdirection)))
     (do (cond ((and (codep (*get-stack i))
		     (wgreaterp (inf (*get-stack i)) heapupperbound))
                (writechar (char tab))
		(channelwriteunknownitem out* (*get-stack i))
                (terpri))
               ((weq (tag (getmem i)) btr-tag)
		(printf "       %r%n" (mkid (inf (*get-stack i)))))
	       ((setq x (returnaddressp (*get-stack i)))
                (printf "%p -> %p:%n"
			(code-address-to-symbol (inf (*get-stack i))) x))
	       (t (printf "     %p%n" (*get-stack i))))))
    (return (terpri))))

(de backtrace1 (item code)
  %
  % Code is 1 if Interpreter functions should be printed, 0 if not.
  %
  (unless (memq item ignoredinbacktrace*)
    (unless (and (equal code 0) (memq item interpreterfunctions*))
      (prin1 item)
      (writechar (char blank)))))

(off fast-integers)
